﻿using System;
using System.Threading;

namespace InteractivePulseAndWait {
	class Program {
		static void ThreadBody(object communicationObject) {
			while (true) {
				Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: Starting next infinite cycle iteration ...");
				lock (communicationObject) {
					Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: Implicit Monitor.Enter() successful -> entering Wait() ...");
					Monitor.Wait(communicationObject);
				}
				Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: SOMEONE PULSED ME, entering Sleep() ...");
				Thread.Sleep(3000);
				Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: WOKE UP from Sleep() ...");
			}
		}

		static void Main(string[] args) {
			var communicationObject1 = new object();
			var communicationObject2 = new object();

			var t1 = new Thread(ThreadBody);
			t1.IsBackground = true;
			t1.Start(communicationObject1);

			var t2 = new Thread(ThreadBody);
			t2.IsBackground = true;
			t2.Start(
#if !SHARED_OBJECT
				communicationObject2
#else   
				communicationObject1
#endif
			);

			Console.WriteLine($"*** Main thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: Entering interactive mode (Enter = pulse @ 1, Space = pulse @ 2, A = pulse all @ 1, Escape = exit) ...");
			while (true) {
				var key = Console.ReadKey(intercept: true);
				Console.WriteLine($"*** Main thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: Key {key.Key} press detected ...");
				if (key.Key == ConsoleKey.Enter) {
					lock (communicationObject1) {
						Monitor.Pulse(communicationObject1);
					}
				} else if (key.Key == ConsoleKey.Spacebar) {
					lock (communicationObject2) {
						Monitor.Pulse(communicationObject2);
					}
				} else if (key.Key == ConsoleKey.A) {
					lock (communicationObject1) {
						Monitor.PulseAll(communicationObject1);
					}
				} else if (key.Key == ConsoleKey.Escape) {
					break;
				}
			}
			Console.WriteLine($"*** Main thread {Thread.CurrentThread.ManagedThreadId} @ CPU {Thread.GetCurrentProcessorId()}: Terminating ...");
		}
	}
}
