﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace ThreadStaticVsThreadLocal {
	class X {
		public ThreadLocal<int> x = new ThreadLocal<int>(() => 10);
		public ThreadLocal<int> y = new ThreadLocal<int>(() => 20);
	}

	class Program {
		[ThreadStatic]
		static int a = 42;

		static ThreadLocal<int> b = new ThreadLocal<int>(() => 42);

		static void Main(string[] args) {
			Console.WriteLine("Thread #{0}: Main starts ...", Thread.CurrentThread.ManagedThreadId); Console.WriteLine();

			X x1 = new X();
			X x2 = new X();
			
			x1.x.Value++;
			x2.y.Value++;
			Console.WriteLine("Thread #{0}: ***** x1.x++, x2.y++", Thread.CurrentThread.ManagedThreadId); Console.WriteLine();

			ThreadStart printAll = () => {
				Console.WriteLine("Thread #{0}: printAll is starting ...", Thread.CurrentThread.ManagedThreadId);

				Console.WriteLine("Thread #{1}: a == {0}", a, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: b == {0}", b.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x1.x == {0}", x1.x.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x1.y == {0}", x1.y.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x2.x == {0}", x2.x.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x2.y == {0}", x2.y.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{0}: ***** x1.x += 100, x2.y += 100, a += 100, b += 100", Thread.CurrentThread.ManagedThreadId);

				x1.x.Value += 100;
				x2.y.Value += 100;
				a += 100;
				b.Value += 100;

				Console.WriteLine("Thread #{1}: a == {0}", a, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: b == {0}", b.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x1.x == {0}", x1.x.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x1.y == {0}", x1.y.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x2.x == {0}", x2.x.Value, Thread.CurrentThread.ManagedThreadId);
				Console.WriteLine("Thread #{1}: x2.y == {0}", x2.y.Value, Thread.CurrentThread.ManagedThreadId);

				Console.WriteLine("Thread #{0}: printAll is ending ...", Thread.CurrentThread.ManagedThreadId);
			};

			var t = new Thread(printAll);
			t.Start();
			t.Join();

			Console.WriteLine();
			Console.WriteLine("Thread #{0}: Main calling printAll ...", Thread.CurrentThread.ManagedThreadId);
			printAll();
			Console.WriteLine();

			List<X> xes = new List<X>();
			for (int i = 0; i < 10000000; i++) {
				xes.Add(new X());
			}
			Console.WriteLine(xes.Count);

			Console.WriteLine("Thread #{0}: Main ends ...", Thread.CurrentThread.ManagedThreadId); 
		}
	}
}
