﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

using System.Threading;

namespace WpfAppSynchronizationContext {
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window {

		public MainWindow() : this("") {    // Explicit parameterless constructor is necessary, as XAML parser is not able to use MainWindow(string) with default parameter value instead of it.
		}

		public MainWindow(string initialLog) {
			InitializeComponent();

			logTextBlock.Text += initialLog;
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: MainWindow(): SynchronizationContext.Current == {SynchronizationContext.Current}, "
				+ $"Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";
		}

		private void blockButton_Click(object sender, RoutedEventArgs e) {
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: blockButton_Click() before Sleep(): SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";
			Thread.Sleep(5000);
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: blockButton_Click() after Sleep(): SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";
		}

		private void createWindowInThreadButton_Click(object sender, RoutedEventArgs e) {
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: createWindowInThreadButton_Click";

			var t = new Thread(() => {
				var log = $"\nThread #{Thread.CurrentThread.ManagedThreadId}: before new MainWindow(): SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher?.GetHashCode()}";

				var window = new MainWindow(log);
				window.Closed += (_, _) => {
					System.Diagnostics.Debug.WriteLine($"Thread #{Thread.CurrentThread.ManagedThreadId}: Windows.Closed -> Shuting down Dispatcher ...");
					// Exit current Dispatcher when Window closes
					System.Windows.Threading.Dispatcher.CurrentDispatcher!.BeginInvokeShutdown(System.Windows.Threading.DispatcherPriority.SystemIdle);
					System.Windows.Threading.Dispatcher.Run();
					System.Diagnostics.Debug.WriteLine($"Thread #{Thread.CurrentThread.ManagedThreadId}: Windows.Closed -> Dispatcher is terminated ...");
				};

				window.logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: after new MainWindow(): SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher?.GetHashCode()}";

				window.ShowDialog();
			});
			t.SetApartmentState(ApartmentState.STA);
			t.Start();

			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: createWindowInThreadButton_Click created thread #{t.ManagedThreadId}";
		}

		private void directUpdateButton_Click(object sender, RoutedEventArgs e) {
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: directUpdateButton_Click: SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";

			var t = new Thread(() => {
				Thread.Sleep(1000);
				logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: DONE processing ...";
			});
			t.Start();

			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: directUpdateButton_Click created thread #{t.ManagedThreadId}";
		}

		private void dispatcherUpdateButton_Click(object sender, RoutedEventArgs e) {
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: dispatcherUpdateButton_Click: SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";

			var t = new Thread(() => {
				Thread.Sleep(1000);
				var message = $"Thread #{Thread.CurrentThread.ManagedThreadId}: DONE processing ...";

				this.Dispatcher.BeginInvoke(() => {
					logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: invoked by Dispatcher, my message: " + message;
				});
			});
			t.Start();

			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: dispatcherUpdateButton_Click created thread #{t.ManagedThreadId}";
		}

		private void synchronizationContextUpdateButton_Click(object sender, RoutedEventArgs e) {
			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: synchronizationContextUpdateButton_Click: SynchronizationContext.Current == {SynchronizationContext.Current}, Dispatcher == {System.Windows.Threading.Dispatcher.CurrentDispatcher.GetHashCode()}";

			var capturedContext = SynchronizationContext.Current;

			var t = new Thread(() => {
				Thread.Sleep(1000);
				var message = $"Thread #{Thread.CurrentThread.ManagedThreadId}: DONE processing ..., SynchronizationContext.Current == {SynchronizationContext.Current}";

				capturedContext!.Post(_ => {
					logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: invoked by SynchronizationContext, my message: " + message;
				}, null);
			});
			t.Start();

			logTextBlock.Text += $"\nThread #{Thread.CurrentThread.ManagedThreadId}: synchronizationContextUpdateButton_Click created thread #{t.ManagedThreadId}";
		}
	}
}
