﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using System.Threading;

namespace WindowsFormsThreadingModel {
	public partial class AppForm : Form {
		private AppForm parentForm = null;

		public AppForm() {
			InitializeComponent();		
		}

		protected override void OnLoad(EventArgs e) {
			base.OnLoad(e);

			Text = Text + string.Format(": {0}, Parent: '{1}'", Thread.CurrentThread.ManagedThreadId, parentForm);
		}

		private void showInfoButton_Click(object sender, EventArgs e) {
			MessageBox.Show(string.Format("ManagedThreadId: {0}", Thread.CurrentThread.ManagedThreadId));
		}

		private void addFormHereButton_Click(object sender, EventArgs e) {
			var f = new AppForm();
			f.parentForm = this;
			f.Show();
		}

		private void addFormInThreadButton_Click(object sender, EventArgs e) {
			var t = new Thread(() => {
				var f = new AppForm();
				f.parentForm = this;
				Application.Run(f);
			});
			t.Start();
		}

		private void setLabelButton_Click(object sender, EventArgs e) {
			label1.Text = DateTime.Now.ToLongTimeString();
		}

		private void setLabelInThreadButton_Click(object sender, EventArgs e) {
			var t = new Thread(() => {
				label1.Text = DateTime.Now.ToLongTimeString();
			});
			t.Start();
		}

		private void setLabelInParentButton_Click(object sender, EventArgs e) {
			if (parentForm == null) {
				MessageBox.Show("No parent form assigned.");
			} else {
				parentForm.label1.Text = DateTime.Now.ToLongTimeString();
			}
		}

		private void setLabelViaInvokeButton_Click(object sender, EventArgs e) {
			var t = new Thread(() => {

				this.Invoke((Action) (
					() => {
						label1.Text = DateTime.Now.ToLongTimeString();
					}
				));
				
			});
			t.Start();
		}

		private void setParentLabelViaInvokeButton_Click(object sender, EventArgs e) {
			if (parentForm == null) {
				MessageBox.Show("No parent form assigned.");
			} else {
				if (parentForm.InvokeRequired) {

					parentForm.Invoke((Action) (
						() => {
							parentForm.label1.Text = DateTime.Now.ToLongTimeString() + " via Invoke";
						}
					));

				} else {
					parentForm.label1.Text = DateTime.Now.ToLongTimeString() + " directly set";
				}
			}
		}
	}
}
