﻿
var ctx = new CustomQueueSynchronizationContext();
SynchronizationContext.SetSynchronizationContext(ctx);
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 1A");
var future1a = GetValueAsync(1, 2);
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 1B");
var future1b = GetValueAsync(3, 4);
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 1C");
ctx.RunJobExectutionLoop();
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 1D");

Console.WriteLine();
Console.WriteLine("Press ENTER to start example 2 ...");
Console.ReadLine();

Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 2A");
var future2 = GetValueAsync_v2(5, 6);
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 2B");
ctx.RunJobExectutionLoop();
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 2C");
future2.Wait();
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 2D");

Console.WriteLine();
Console.WriteLine("Press ENTER to start example 3 ...");
Console.ReadLine();

Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 3A");
var future3 = GetValueAsync_v3(7, 8);
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 3B");
ctx.RunJobExectutionLoop();
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 3C");
future3.Wait();
Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: Main 3D");

static async Task<int> GetValueAsync(int a, int b) {
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync step 0");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync step 1");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync step 2");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync step 3");
	return a + b;
}

static async Task<int> GetValueAsync_v2(int a, int b) {
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v2 step 0");
	await Task.Delay(1000).ConfigureAwait(continueOnCapturedContext: false);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v2 step 1");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v2 step 2");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v2 step 3");
	return a + b;
}

static async Task<int> GetValueAsync_v3(int a, int b) {
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v3 step 0");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v3 step 1");
	await Task.Delay(1000).ConfigureAwait(continueOnCapturedContext: false);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v3 step 2");
	await Task.Delay(1000);
	Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: GetValueAsync_v3 step 3");
	return a + b;
}

class CustomQueueSynchronizationContext : SynchronizationContext {
	private readonly Queue<(SendOrPostCallback, object?)> _jobQueue = new Queue<(SendOrPostCallback, object?)>();

	public override void Post(SendOrPostCallback d, object? state) {
		string info = $"Thread {Thread.CurrentThread.ManagedThreadId}: POST before call - d.Target.GetHashCode() == {d.Target!.GetHashCode()}";
		if (state is MulticastDelegate md) {
			info += $"  state.Target.GetHashCode() == {md.Target!.GetHashCode()} ";
		}
        Console.WriteLine(info);
        lock (_jobQueue) {
			_jobQueue.Enqueue((d, state));
			Monitor.Pulse(_jobQueue);
		}
	}

	public void RunJobExectutionLoop() {
		Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: STARTING JOB LOOP ...");
		while (true) {
			SendOrPostCallback jobDelegate;
			object? state;

			lock (_jobQueue) {
				if (_jobQueue.Count == 0) {
					if (!Monitor.Wait(_jobQueue, millisecondsTimeout: 2000)) {
						break;
					}
				}

				(jobDelegate, state) = _jobQueue.Dequeue();
			}

			string info = $"Thread {Thread.CurrentThread.ManagedThreadId}: JOB LOOP before call - d.Target.GetHashCode() == {jobDelegate.Target!.GetHashCode()}";
			if (state is MulticastDelegate md) {
				info += $"  state.Target.GetHashCode() == {md.Target!.GetHashCode()} ";
			}
			Console.WriteLine(info);
			jobDelegate(state);
			Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: JOB LOOP after call");
		}
		Console.WriteLine($"Thread {Thread.CurrentThread.ManagedThreadId}: ENDING JOB LOOP ...");
	}
}