﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace TopStoriesConfigureAwaitWinForms {
	public partial class Form1 : Form {
		public Form1() {
			InitializeComponent();
		}

		private static readonly string[] AnimationSymbols = { "-", "\\", "|", "/" };

		private int animationStep = 0;

		private void animationTimer_Tick(object sender, EventArgs e) {
			animationLabel.Text = AnimationSymbols[animationStep];
			animationStep = (animationStep + 1) % AnimationSymbols.Length;
		}

		private Model model = new Model();

		private async void loadTopStoriesButton_Click(object sender, EventArgs e) {
			Console.WriteLine("-> loadTopStoriesButton_Click handler starts ...");
			loadTopStoriesButton.Enabled = false;
			animationTimer.Enabled = true;
			storyListBox.Items.Clear();

			var topStories = await model.GetTopStoriesAsync(5);
			// var topStories = await Task.Run(() => model.GetTopStories(5));

			foreach (var story in topStories) {
				storyListBox.Items.Add(story.Headline);
			}

			animationTimer.Enabled = false;
			loadTopStoriesButton.Enabled = true;
			Console.WriteLine("<- loadTopStoriesButton_Click handler ends ...");
		}
	}

	public class Model {
		public async Task<IReadOnlyList<Story>> GetTopStoriesAsync(int storyCountLimit) {
			Console.WriteLine("-> GetTopStoriesAsync called ...");
			var topStoryList = new List<Story>();

			var topStoryIDs = await GetTopStoryIDsAsync(); // .ConfigureAwait(continueOnCapturedContext: false);
			storyCountLimit = Math.Min(topStoryIDs.Count, storyCountLimit);
			for (int i = 0; i < storyCountLimit; i++) {
				var story = await GetStoryAsync(topStoryIDs[i]);
				Console.WriteLine("++ GetTopStoriesAsync adding new story to List ...");
				topStoryList.Add(story);
			}

			Console.WriteLine("<- GetTopStoriesAsync done ...");
			return topStoryList;
		}

		public async Task<IReadOnlyList<string>> GetTopStoryIDsAsync() {
			Console.WriteLine("  -> GetTopStoryIDsAsync before await ...");
			await Task.Delay(2000); // .ConfigureAwait(continueOnCapturedContext: false);
			Console.WriteLine("  <- GetTopStoryIDsAsync after await ...");
			return new List<string> { "A001", "B001", "B002" /*, "FAIL" */};
		}

		public async Task<Story> GetStoryAsync(string storyId) {
			Console.WriteLine("  -> GetStoryAsync before await and processing ...");
			await Task.Delay(1000);
			// var result = stories.FirstOrDefault(story => story.ID == storyId);
			var result = stories.First(story => story.ID == storyId);
			Console.WriteLine("  <- GetStoryAsync after await and processing ...");
			return result;
		}

		private static readonly List<Story> stories = new List<Story> {
			new Story(id: "A001", headline: "Coronavirus entered Czech Republic."),
			new Story(id: "A002", headline: "Vltava river opened its secrets."),
			new Story(id: "A003", headline: "Rats do not eat sugar."),
			new Story(id: "B001", headline: "Trump flew around Brno."),
			new Story(id: "B002", headline: "Two newborn wolves in Prague ZOO.")
		};

		public class Story {
			public Story(string id, string headline) {
				ID = id;
				Headline = headline;
			}

			public string ID { get; }
			public string Headline { get; }
		}
	}

	class Console {
		public static void WriteLine(string text) {
			System.Console.WriteLine($"Thread #{System.Threading.Thread.CurrentThread.ManagedThreadId}: {text}");
		}
	}
}
